#include "stdlib.h"
#include "stdio.h"
#include "TEncAuf.h"
#include "math.h"
#include "assert.h"
#include "TLibCommon/TComPicYuv.h"

#if TE4_4_2_1

Int usf_luma_fixed[16][NTAPS_US_LUMA] =
{
  {  0,  0,  0, 64,  0,  0,  0,  0}, //
  { -1, -1, -1, -1, -1, -1, -1, -1}, //
  { -1, -1, -1, -1, -1, -1, -1, -1}, //
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, //
  { -1, 4, -11, 52, 26,  -8, 3, -1}, // <-> actual phase shift 1/3, used for spatial scalability x1.5      
  { -1, -1, -1, -1, -1, -1, -1, -1}, //       
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, 4, -11, 40, 40, -11, 4, -1}, // <-> actual phase shift 1/2, equal to HEVC MC, used for spatial scalability x2
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, 3,  -8, 26, 52, -11, 4, -1}, // <-> actual phase shift 2/3, used for spatial scalability x1.5
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}  // 
};

#ifndef CNU
#define CNU -1
#endif
Int usf_chroma_fixed[16][NTAPS_US_CHROMA] =
{
  {  0, 64,  0,  0},//
  {CNU,CNU,CNU,CNU},//
  {CNU,CNU,CNU,CNU},//
  {CNU,CNU,CNU,CNU},// 
  { -4, 54, 16, -2},// <-> actual phase shift 1/4,equal to HEVC MC, used for spatial scalability x1.5 (only for accurate Chroma alignement)
  { -5, 50, 22, -3},// <-> actual phase shift 1/3, used for spatial scalability x1.5   
  { -6, 46, 28, -4},// <-> actual phase shift 3/8,equal to HEVC MC, used for spatial scalability x2 (only for accurate Chroma alignement)      
  {CNU,CNU,CNU,CNU},// 
  { -4, 36, 36, -4},// <-> actual phase shift 1/2,equal to HEVC MC, used for spatial scalability x2
  { -4, 30, 43, -5},// <-> actual phase shift 7/12, used for spatial scalability x1.5 (only for accurate Chroma alignement)
  {CNU,CNU,CNU,CNU},// 
  { -3, 22, 50, -5},// <-> actual phase shift 2/3, used for spatial scalability x1.5
  {CNU,CNU,CNU,CNU},// 
  {CNU,CNU,CNU,CNU},// 
  { -2, 10, 58, -2},// <-> actual phase shift 7/8,equal to HEVC MC, used for spatial scalability x2 (only for accurate Chroma alignement)  
  { -1,  5, 62, -2} // <-> actual phase shift 11/12, used for spatial scalability x1.5 (only for accurate Chroma alignement)
};

TEncAuf::TEncAuf(void)
{
  m_pPicYuvTemp[0] = NULL;
  m_pPicYuvTemp[1] = NULL;
  m_pPicYuvTemp[2] = NULL;
}

TEncAuf::~TEncAuf(void)
{
  if ( m_pPicYuvTemp[0] )
  {
#if IL_AUF
    m_pPicYuvTemp[0]->destroy();
#else
    m_pPicYuvTemp[0]->destroyLuma();
#endif
    delete m_pPicYuvTemp[0];
    m_pPicYuvTemp[0] = NULL;
  }

  if ( m_pPicYuvTemp[1] )
  {
#if IL_AUF
    m_pPicYuvTemp[1]->destroy();
#else
    m_pPicYuvTemp[1]->destroyLuma();
#endif
    delete m_pPicYuvTemp[1];
    m_pPicYuvTemp[1] = NULL;
  }

  if (m_pPicYuvTemp[2])
  {
    m_pPicYuvTemp[2]->destroy();
    delete m_pPicYuvTemp[2];
    m_pPicYuvTemp[2] = NULL;
  }
}

Int TEncAuf::computeAufFilters(TComPicYuv* pPicYuvOrg, TComPicYuv* pPicYuvRecBase, AUFParam* aufParam, Int iSliceInterval, TComPicYuv* pPicMap,
                               Double dLambdaLuma, Double dLambdaChroma)
{
  Pel* pOrgY   = pPicYuvOrg->getLumaAddr();
  Int orgWidth = pPicYuvOrg->getWidth() - pPicYuvOrg->getPicCropLeftOffset() - pPicYuvOrg->getPicCropRightOffset();
  Int orgHeight = pPicYuvOrg->getHeight() - pPicYuvOrg->getPicCropTopOffset() - pPicYuvOrg->getPicCropBottomOffset();
  Int orgStride = pPicYuvOrg->getStride();
  Pel* pRefY    = pPicYuvRecBase->getLumaAddr();
  Int refWidth  = pPicYuvRecBase->getWidth() - pPicYuvRecBase->getPicCropLeftOffset() - pPicYuvRecBase->getPicCropRightOffset();
  Int refHeight = pPicYuvRecBase->getHeight() - pPicYuvRecBase->getPicCropTopOffset() - pPicYuvRecBase->getPicCropBottomOffset();
  Int refStride = pPicYuvRecBase->getStride();

  m_interWidth = refWidth;
  m_interHeight = orgHeight;
  Int dir; // hor/ver;
  Int iSpatialIdx;

#if IL_AUF
  Pel *pOrgU = pPicYuvOrg->getCbAddr();
  Pel *pOrgV = pPicYuvOrg->getCrAddr();
  Int orgCWidth = orgWidth >> 1;
  Int orgCHeight = orgHeight >> 1;
  Int orgCStride = pPicYuvOrg->getCStride();

  Pel *pRefU = pPicYuvRecBase->getCbAddr();
  Pel *pRefV = pPicYuvRecBase->getCrAddr();
  Int refCWidth = refWidth >> 1;
  Int refCHeight = refHeight >> 1;
  Int refCStride = pPicYuvRecBase->getCStride();
#endif

  if (orgWidth ==  refWidth)
  {
    iSpatialIdx = 0;
  }
  else if (refWidth*3 ==  orgWidth*2)
  {
    iSpatialIdx = 1;
  }
  else if (refWidth*2 ==  orgWidth)
  {
    iSpatialIdx = 2;
  }
  else 
  {
    assert(0);
  }

#if IL_AUF
  aufParam->setSpatialIdx(COLOR_Y, iSpatialIdx);
  aufParam->setSpatialIdx(COLOR_Cb, iSpatialIdx);
  aufParam->setSpatialIdx(COLOR_Cr, iSpatialIdx);
#else
  aufParam->setSpatialIdx (iSpatialIdx);
#endif

  if (m_pPicYuvTemp[0] == NULL)
  {
    m_pPicYuvTemp[0] = new TComPicYuv;
#if IL_AUF
    m_pPicYuvTemp[0]->create(orgWidth, orgHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);  
#else
    m_pPicYuvTemp[0]->createLuma( orgWidth, orgHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);  
#endif
  }

  if (m_pPicYuvTemp[1] == NULL)
  {
    m_pPicYuvTemp[1] = new TComPicYuv;
#if IL_AUF
    m_pPicYuvTemp[1]->create(orgWidth, orgHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);  
#else
    m_pPicYuvTemp[1]->createLuma( orgWidth, orgHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);  
#endif
  }

#if IL_AUF
  if (m_pPicYuvTemp[2] == NULL)
  {
    m_pPicYuvTemp[2] = new TComPicYuv;
    m_pPicYuvTemp[2]->create(orgWidth, orgHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);  
  }
#endif
  if ( refWidth != orgWidth )
  {
    pPicYuvOrg->copyToPicLuma(m_pPicYuvTemp[1]);
#if IL_AUF
    pPicYuvOrg->copyToPicCb(m_pPicYuvTemp[1]);
    pPicYuvOrg->copyToPicCr(m_pPicYuvTemp[1]);
#endif
  }

  dir = m_interHeight != orgHeight;
  if ( refWidth != orgWidth )
  {
    xDownsampleOneD(m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), m_pPicYuvTemp[1]->getLumaAddr(), orgWidth, orgHeight, m_pPicYuvTemp[1]->getStride(), dir );
#if IL_AUF
    xDownsampleOneD(m_pPicYuvTemp[0]->getCbAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), m_pPicYuvTemp[1]->getCbAddr(), orgCWidth, orgCHeight, m_pPicYuvTemp[1]->getCStride(), dir);
    xDownsampleOneD(m_pPicYuvTemp[0]->getCrAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), m_pPicYuvTemp[1]->getCrAddr(), orgCWidth, orgCHeight, m_pPicYuvTemp[1]->getCStride(), dir);
    
    if (pPicMap)
    {
      pPicMap->copyToPic(m_pPicYuvTemp[2]);
      xDownsampleOneD(pPicMap->getLumaAddr(), m_interWidth, m_interHeight, pPicMap->getStride(), 
                      m_pPicYuvTemp[2]->getLumaAddr(), orgWidth, orgHeight, m_pPicYuvTemp[2]->getStride(), dir);
      xDownsampleOneD(pPicMap->getCbAddr(), m_interWidth >> 1, m_interHeight >> 1, pPicMap->getCStride(), 
                      m_pPicYuvTemp[2]->getCbAddr(), orgCWidth, orgCHeight, m_pPicYuvTemp[2]->getCStride(), dir);
      xDownsampleOneD(pPicMap->getCrAddr(), m_interWidth >> 1, m_interHeight >> 1, pPicMap->getCStride(), 
                      m_pPicYuvTemp[2]->getCrAddr(), orgCWidth, orgCHeight, m_pPicYuvTemp[2]->getCStride(), dir);
    }
#endif
  }

  dir = !dir;
  if ( refWidth == orgWidth )
  {
#if IL_AUF
    xGetStatistics(pOrgY, orgWidth, orgHeight, orgStride, pRefY, refWidth, refHeight, refStride, dir, COLOR_Y, pPicMap);
    xGetStatistics(pOrgU, orgCWidth, orgCHeight, orgCStride, pRefU, refCWidth, refCHeight, refCStride, dir, COLOR_Cb, pPicMap);
    xGetStatistics(pOrgV, orgCWidth, orgCHeight, orgCStride, pRefV, refCWidth, refCHeight, refCStride, dir, COLOR_Cr, pPicMap);
#else
    xGetStatistics( pOrgY, orgWidth, orgHeight, orgStride, pRefY, refWidth, refHeight, refStride, dir);
#endif
  }
  else
  {
#if IL_AUF
    xGetStatistics(m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), pRefY, refWidth, refHeight, refStride, dir, COLOR_Y, pPicMap);
    xGetStatistics(m_pPicYuvTemp[0]->getCbAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), pRefU, refCWidth, refCHeight, refCStride, dir, COLOR_Cb, pPicMap);
    xGetStatistics(m_pPicYuvTemp[0]->getCrAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), pRefV, refCWidth, refCHeight, refCStride, dir, COLOR_Cr, pPicMap);
#else
    xGetStatistics( m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), pRefY, refWidth, refHeight, refStride, dir);
#endif
  }

  xDeriveFilterCoef(dir);
  xQuantizeCoeff(dir, iSliceInterval); 

#if IL_AUF
  xUpsampleOneD(m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), pRefY, refWidth, refHeight, refStride, dir, COLOR_Y);
  xUpsampleOneD(m_pPicYuvTemp[0]->getCbAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), pRefU, refCWidth, refCHeight, refCStride, dir, COLOR_Cb);
  xUpsampleOneD(m_pPicYuvTemp[0]->getCrAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), pRefV, refCWidth, refCHeight, refCStride, dir, COLOR_Cr);
#else
  xUpsampleOneD(m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), pRefY, refWidth, refHeight, refStride, dir);
#endif

  dir = !dir;
  if ( refWidth == orgWidth )
  {
#if IL_AUF
    xGetStatistics(pOrgY, orgWidth, orgHeight, orgStride,  m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), dir, COLOR_Y, pPicMap);
    xGetStatistics(pOrgU, orgCWidth, orgCHeight, orgCStride, m_pPicYuvTemp[0]->getCbAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), dir, COLOR_Cb, pPicMap);
    xGetStatistics(pOrgV, orgCWidth, orgCHeight, orgCStride, m_pPicYuvTemp[0]->getCrAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), dir, COLOR_Cr, pPicMap);
#else
    xGetStatistics( pOrgY, orgWidth, orgHeight, orgStride,  m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), dir);
#endif
  }
  else
  {
#if IL_AUF
    if (pPicMap)
    {
      m_pPicYuvTemp[2]->copyToPic(pPicMap);
    }
    xGetStatistics(m_pPicYuvTemp[1]->getLumaAddr(), orgWidth, orgHeight, m_pPicYuvTemp[1]->getStride(), m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), dir, COLOR_Y, pPicMap);
    xGetStatistics(m_pPicYuvTemp[1]->getCbAddr(), orgCWidth, orgCHeight, m_pPicYuvTemp[1]->getCStride(), m_pPicYuvTemp[0]->getCbAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), dir, COLOR_Cb, pPicMap);
    xGetStatistics(m_pPicYuvTemp[1]->getCrAddr(), orgCWidth, orgCHeight, m_pPicYuvTemp[1]->getCStride(), m_pPicYuvTemp[0]->getCrAddr(), m_interWidth >> 1, m_interHeight >> 1, m_pPicYuvTemp[0]->getCStride(), dir, COLOR_Cr, pPicMap);
#else
    xGetStatistics( m_pPicYuvTemp[1]->getLumaAddr(), orgWidth, orgHeight, m_pPicYuvTemp[1]->getStride(), m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), dir);
#endif
  }

  xDeriveFilterCoef(dir);

  xQuantizeCoeff(dir, iSliceInterval);

#if IL_AUF
  if (xSelectFilter(pPicYuvOrg, pPicYuvRecBase, COLOR_Y, pPicMap, dLambdaLuma, aufParam->getAufCoeffNum(COLOR_Y) * 3 + 10))
  {
    aufParam->setAufFlag(COLOR_Y, true);
#else
  if (xSelectFilter(pPicYuvOrg, pPicYuvRecBase))
  {
    aufParam->setAufFlag (true);
#endif
    for( Int i = 0; i < 2; i++)
    {
      for( Int j = 0; j < 16; j++)
      {
#if IL_AUF
        ::memcpy(aufParam->getAufCoeff(COLOR_Y, i, j), &m_iCoeff[COLOR_Y][i][j][0], sizeof(Int) * NTAPS_US_LUMA);
#else
        memcpy( aufParam->getAufCoeff( i, j), &m_iCoeff[i][j][0], sizeof(Int) * NTAPS_US_LUMA );
#endif
      }
    }
  }
  else
  {
#if IL_AUF
    aufParam->setAufFlag(COLOR_Y, false);
#else
    aufParam->setAufFlag (false);
#endif
  }

#if IL_AUF
  // Cb adaptive filter
  if (xSelectFilter(pPicYuvOrg, pPicYuvRecBase, COLOR_Cb, pPicMap, dLambdaChroma, aufParam->getAufCoeffNum(COLOR_Cb) * 3 + 10))
  {
    aufParam->setAufFlag(COLOR_Cb, true);
    for (Int i = 0; i < 2; i++)
    {
      for (Int j = 0; j < 16; j++)
      {
        ::memcpy(aufParam->getAufCoeff(COLOR_Cb, i, j), &m_iCoeff[COLOR_Cb][i][j][0], sizeof(Int) * NTAPS_US_CHROMA);
      }
    }
  }
  else
  {
    aufParam->setAufFlag(COLOR_Cb, false);
  }

  // Cr adaptive filter
  if (xSelectFilter(pPicYuvOrg, pPicYuvRecBase, COLOR_Cr, pPicMap, dLambdaChroma, aufParam->getAufCoeffNum(COLOR_Cr) * 3 + 10))
  {
    aufParam->setAufFlag(COLOR_Cr, true);
    for (Int i = 0; i < 2; i++)
    {
      for (Int j = 0; j < 16; j++)
      {
        ::memcpy(aufParam->getAufCoeff(COLOR_Cr, i, j), &m_iCoeff[COLOR_Cr][i][j][0], sizeof(Int) * NTAPS_US_CHROMA);
      }
    }
  }
  else
  {
    aufParam->setAufFlag(COLOR_Cr, false);
  }
#endif
  aufParam->encAufPara();

  return 1;
}

#if IL_AUF
Void TEncAuf::xDeriveFilterCoef(Int dir)
{
  for (Int iColor = 0; iColor < 3; iColor++)
  {
    for (Int filterIdx = 0 ; filterIdx < FILTER_SET; filterIdx++)
    {
      if (m_iCount[iColor][dir][filterIdx])
      {
        if (iColor == COLOR_Y)
        {
          gnsSolveByChol(m_dAutoCorr[iColor][dir][filterIdx], m_dCrosCorr[iColor][dir][filterIdx], m_dCoeff[iColor][dir][filterIdx], NTAPS_US_LUMA);
        }
        else
        {
          gnsSolveByChol(m_dAutoCorr[iColor][dir][filterIdx], m_dCrosCorr[iColor][dir][filterIdx], m_dCoeff[iColor][dir][filterIdx], NTAPS_US_CHROMA);
        }
      }
    }
  }
}

void TEncAuf::xQuantizeCoeff(Int dir, Int iSliceInterval)
{
  Int filterIdx;
  Int iUpsampleFilterTaps;

  for (Int iColor = 0; iColor < 3; iColor++)
  {
    iUpsampleFilterTaps = (iColor == COLOR_Y) ? NTAPS_US_LUMA : NTAPS_US_CHROMA;
    for (filterIdx = 1; filterIdx < 8; filterIdx++)
    {
      if (m_iCount[iColor][dir][filterIdx])
      {
        Int sum = 0;
        for (Int i = 0; i < iUpsampleFilterTaps; i++)
        {
          if (i == iUpsampleFilterTaps / 2 - 1 + (filterIdx % 16 + 7) / 16)
          {
            continue;
          }
          Double sign = m_dCoeff[iColor][dir][filterIdx][i] > 0.0 ? 1.0: -1.0;
          m_dCoeff[iColor][dir][filterIdx][i] = ((Int)(m_dCoeff[iColor][dir][filterIdx][i] * (1 << US_FILTER_PREC) + 0.5 * sign)) / (Double)(1 << US_FILTER_PREC);

          m_iCoeff[iColor][dir][filterIdx][i] = ((Int)(m_dCoeff[iColor][dir][filterIdx][i] * (1 << US_FILTER_PREC) + 0.5 * sign));
          sum +=  m_iCoeff[iColor][dir][filterIdx][i];
        }
        m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1 + (filterIdx % 16 + 7) / 16] = (1 << US_FILTER_PREC) - sum;
        m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1 + (filterIdx % 16 + 7) / 16] = m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1 + (filterIdx%16+7)/16] / (double)(1 << US_FILTER_PREC);
      }
      else
      {
        for (Int i = 0; i < iUpsampleFilterTaps; i++)
        {
          m_iCoeff[iColor][dir][filterIdx][i] = (iColor == COLOR_Y) ? usf_luma_fixed[filterIdx][i] : usf_chroma_fixed[filterIdx][i];
        }
      }
    }

    for (filterIdx = 1 ; filterIdx < 8; filterIdx++)
    {
      for (Int i = 0; i < iUpsampleFilterTaps; i++)
      {
        m_iCoeff[iColor][dir][16 - filterIdx][i] = m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1 - i];
        m_dCoeff[iColor][dir][16 - filterIdx][i] = m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1 - i];
      }
    }

    filterIdx = 0;
    if (m_iCount[iColor][dir][filterIdx])
    {
      Int sum = 0;
      m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1] = m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1] = 0;
      for (Int i = 0; i < iUpsampleFilterTaps / 2 - 1; i++ )
      {
        Double sign = (m_dCoeff[iColor][dir][filterIdx][i] + m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 2 - i]) > 0.0 ? 1.0: -1.0;
        m_dCoeff[iColor][dir][filterIdx][i] = ((Int)((m_dCoeff[iColor][dir][filterIdx][i] + m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 2 - i] ) * (1 << US_FILTER_PREC) / 2.0 + 0.5 * sign)) / (Double)(1 << US_FILTER_PREC);
        m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 2 - i] = m_dCoeff[iColor][dir][filterIdx][i];

        m_iCoeff[iColor][dir][filterIdx][i] = ((Int)(m_dCoeff[iColor][dir][filterIdx][i] * (1 << US_FILTER_PREC) + 0.5 * sign));
        m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 2 - i] = m_iCoeff[iColor][dir][filterIdx][i];
        sum += m_iCoeff[iColor][dir][filterIdx][i] * 2;
      }
      m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1] = (1 << US_FILTER_PREC) - sum;
      m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1] = m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1] / (Double)(1 << US_FILTER_PREC);
    }
    else
    {
      for (Int i = 0; i < iUpsampleFilterTaps; i++ )
      {
        m_iCoeff[iColor][dir][filterIdx][i] = (iColor == COLOR_Y) ? usf_luma_fixed[filterIdx][i] : usf_chroma_fixed[filterIdx][i];
      }
    }

    filterIdx = 8;
    if (m_iCount[iColor][dir][filterIdx])
    {
      Int sum = 0;
      for (Int i = 0; i < iUpsampleFilterTaps / 2 - 1; i++)
      {
        Double sign = (m_dCoeff[iColor][dir][filterIdx][i] + m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1 - i]) > 0.0 ? 1.0: -1.0;
        m_dCoeff[iColor][dir][filterIdx][i] = ((Int)((m_dCoeff[iColor][dir][filterIdx][i] + m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1 - i]) * (1<<US_FILTER_PREC)/2+0.5*sign)) / (Double)(1 << US_FILTER_PREC);
        m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1 - i] = m_dCoeff[iColor][dir][filterIdx][i];

        m_iCoeff[iColor][dir][filterIdx][i] = ((Int)(m_dCoeff[iColor][dir][filterIdx][i] * (1 << US_FILTER_PREC) + 0.5 * sign));
        m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps - 1 - i] = m_iCoeff[iColor][dir][filterIdx][i];
        sum +=  m_iCoeff[iColor][dir][filterIdx][i] * 2;
      }
      m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1] = ((1<<US_FILTER_PREC) - sum) / 2;
      m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2] = m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1];

      m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1] = m_iCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1] / (Double)(1 << US_FILTER_PREC);
      m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2] = m_dCoeff[iColor][dir][filterIdx][iUpsampleFilterTaps / 2 - 1];

    }
    else
    {
      for (Int i = 0; i < iUpsampleFilterTaps; i++)
      {
        m_iCoeff[iColor][dir][filterIdx][i] = (iColor == COLOR_Y) ? usf_luma_fixed[filterIdx][i] : usf_chroma_fixed[filterIdx][i];
      }
    }

    Int coeff_threshold = 62;
    if (iSliceInterval == 1)
    {
      coeff_threshold += 10;
    }
    if (m_iCoeff[iColor][dir][0][iUpsampleFilterTaps / 2 - 1] >= coeff_threshold) ///< SHOULD have a bug fix here!!!
    {
      for (Int j = 0; j < 16; j++)
      {
        if (iColor == COLOR_Y)
        {
          ::memcpy(&m_iCoeff[iColor][dir][j][0], &usf_luma_fixed[j][0], sizeof(Int) * iUpsampleFilterTaps);
        }
        else
        {
          ::memcpy(&m_iCoeff[iColor][dir][j][0], &usf_chroma_fixed[j][0], sizeof(Int) * iUpsampleFilterTaps);
        }
      }
    }
  }
}

Bool TEncAuf::xSelectFilter(TComPicYuv *pPicYuvOrg, TComPicYuv *pPicYuvRecBase, Int iColor, TComPicYuv *pPicMap, Double dLambda, Int iRate)
{

  Int sumHAD_fix = 0;
  Int sumHAD_auf = 0;

  Pel *pOrgY = NULL;
  Int orgWidth, orgHeight, orgStride;
  Pel *pBaseY = NULL;
  Int baseWidth, baseHeight, baseStride, dstStride;
  Int iMapStride = -1;
  Int iMapWidth = -1;
  Int iMapHeight = -1;
  
  Pel *pDstY = NULL;
  Pel *tempY = NULL;
  Pel *pMap = NULL;
  switch (iColor)
  {
  case COLOR_Y:
    pOrgY = pPicYuvOrg->getLumaAddr();
    orgWidth = pPicYuvOrg->getWidth() - pPicYuvOrg->getPicCropLeftOffset() - pPicYuvOrg->getPicCropRightOffset();
    orgHeight = pPicYuvOrg->getHeight() - pPicYuvOrg->getPicCropTopOffset() - pPicYuvOrg->getPicCropBottomOffset();
    orgStride = pPicYuvOrg->getStride();
    pBaseY    = pPicYuvRecBase->getLumaAddr();
    baseWidth  = pPicYuvRecBase->getWidth() - pPicYuvRecBase->getPicCropLeftOffset() - pPicYuvRecBase->getPicCropRightOffset();
    baseHeight = pPicYuvRecBase->getHeight() - pPicYuvRecBase->getPicCropTopOffset() - pPicYuvRecBase->getPicCropBottomOffset();
    baseStride = pPicYuvRecBase->getStride();

    pDstY = m_pPicYuvTemp[0]->getLumaAddr();
    dstStride = m_pPicYuvTemp[0]->getStride();
    tempY = m_pPicYuvTemp[1]->getLumaAddr();
    
    if (pPicMap)
    {
      pMap = pPicMap->getLumaAddr();
      iMapStride = pPicMap->getStride();
      iMapWidth = pPicMap->getWidth() - pPicMap->getPicCropLeftOffset() - pPicMap->getPicCropRightOffset();
      iMapHeight = pPicMap->getHeight() - pPicMap->getPicCropTopOffset() - pPicMap->getPicCropBottomOffset();
    }
    break;
  case COLOR_Cb:
    pOrgY = pPicYuvOrg->getCbAddr();
    orgWidth = (pPicYuvOrg->getWidth() - pPicYuvOrg->getPicCropLeftOffset() - pPicYuvOrg->getPicCropRightOffset()) >> 1;
    orgHeight = (pPicYuvOrg->getHeight() - pPicYuvOrg->getPicCropTopOffset() - pPicYuvOrg->getPicCropBottomOffset()) >> 1;
    orgStride = pPicYuvOrg->getCStride();
    pBaseY = pPicYuvRecBase->getCbAddr();
    baseWidth = (pPicYuvRecBase->getWidth() - pPicYuvRecBase->getPicCropLeftOffset() - pPicYuvRecBase->getPicCropRightOffset()) >> 1;
    baseHeight = (pPicYuvRecBase->getHeight() - pPicYuvRecBase->getPicCropTopOffset() - pPicYuvRecBase->getPicCropBottomOffset()) >> 1;
    baseStride = pPicYuvRecBase->getCStride();

    pDstY = m_pPicYuvTemp[0]->getCbAddr();
    dstStride = m_pPicYuvTemp[0]->getCStride();
    tempY = m_pPicYuvTemp[1]->getCbAddr();

    if (pPicMap)
    {
      pMap = pPicMap->getCbAddr();
      iMapStride = pPicMap->getCStride();
      iMapWidth = (pPicMap->getWidth() - pPicMap->getPicCropLeftOffset() - pPicMap->getPicCropRightOffset()) >> 1;
      iMapHeight = (pPicMap->getHeight() - pPicMap->getPicCropTopOffset() - pPicMap->getPicCropBottomOffset()) >> 1;
    }
    break;
  case COLOR_Cr:
    pOrgY = pPicYuvOrg->getCrAddr();
    orgWidth = (pPicYuvOrg->getWidth() - pPicYuvOrg->getPicCropLeftOffset() - pPicYuvOrg->getPicCropRightOffset()) >> 1;
    orgHeight = (pPicYuvOrg->getHeight() - pPicYuvOrg->getPicCropTopOffset() - pPicYuvOrg->getPicCropBottomOffset()) >> 1;
    orgStride = pPicYuvOrg->getCStride();
    pBaseY = pPicYuvRecBase->getCrAddr();
    baseWidth = (pPicYuvRecBase->getWidth() - pPicYuvRecBase->getPicCropLeftOffset() - pPicYuvRecBase->getPicCropRightOffset()) >> 1;
    baseHeight = (pPicYuvRecBase->getHeight() - pPicYuvRecBase->getPicCropTopOffset() - pPicYuvRecBase->getPicCropBottomOffset()) >> 1;
    baseStride = pPicYuvRecBase->getCStride();

    pDstY = m_pPicYuvTemp[0]->getCrAddr();
    dstStride = m_pPicYuvTemp[0]->getCStride();
    tempY = m_pPicYuvTemp[1]->getCrAddr();

    if (pPicMap)
    {
      pMap = pPicMap->getCrAddr();
      iMapStride = pPicMap->getCStride();
      iMapWidth = (pPicMap->getWidth() - pPicMap->getPicCropLeftOffset() - pPicMap->getPicCropRightOffset()) >> 1;
      iMapHeight = (pPicMap->getHeight() - pPicMap->getPicCropTopOffset() - pPicMap->getPicCropBottomOffset()) >> 1;
    }
    break;
  default:
    assert(0);
  }

  if (iColor == COLOR_Y)
  {
    xUpsampleTwoD(pDstY, tempY, orgWidth, orgHeight, dstStride, pBaseY, baseWidth, baseHeight, baseStride, usf_luma_fixed, usf_luma_fixed, COLOR_Y);
  }
  else
  {
    xUpsampleTwoDChroma(pDstY, tempY, orgWidth, orgHeight, dstStride, pBaseY, baseWidth, baseHeight, baseStride, usf_chroma_fixed, usf_chroma_fixed);
  }
#if IL_AUF
  sumHAD_fix = m_pcRdCost->calcHAD(pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight, pMap, iMapStride, iMapWidth, iMapHeight);
#else
  sumHAD_fix = m_pcRdCost->calcHAD(pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight);
#endif  
  xUpsampleTwoD(pDstY, tempY, orgWidth, orgHeight, dstStride, pBaseY, baseWidth, baseHeight, baseStride, m_iCoeff[iColor][1], m_iCoeff[iColor][0], iColor);

#if IL_AUF
  sumHAD_auf = m_pcRdCost->calcHAD(pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight, pMap, iMapStride, iMapWidth, iMapHeight);
#else
  sumHAD_auf = m_pcRdCost->calcHAD(pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight);
#endif

  sumHAD_auf += Int(dLambda * iRate);
  if (sumHAD_auf >= sumHAD_fix)
  {
    return false;
  }
  else
  {
    return true;
  }
}

void TEncAuf::xUpsampleTwoDChroma(Pel *pDstY, Pel *tempY, Int dstWidth, Int dstHeight, Int dstStride, Pel *pSrcY, Int srcWidth, Int srcHeight, 
                                  Int srcStride, Int filterVer[16][4], Int filterHor[16][4])
{
  Int iAddX, iScaleX, iAddY, iScaleY;

  Int iShift = 16;
  Int iPhaseX = 0;
  Int iPhaseY = 1;
  Int iShiftM4 = iShift - 4;
  Int iDeltaX = 4 * iPhaseX; 
  Int iDeltaY = 4 * iPhaseY; 
  Int interWidth = srcWidth;
  Int interStride = dstStride;
  Int interHeight = dstHeight;
  Int iTapOffset = NTAPS_US_CHROMA / 2 - 1;

  //===== vertical upsampling =====
  iAddX = (((srcHeight * iPhaseX) << (iShift - 2)) + (interHeight >> 1)) / interHeight + (1 << (iShift - 5));
  iScaleX = ((srcHeight << iShift) + (interHeight >> 1)) / interHeight;

  iAddY = (((srcHeight * iPhaseY) << (iShift - 2)) + (interHeight >> 1)) / interHeight + (1 << (iShift - 5));
  iScaleY = ((srcHeight << iShift) + (interHeight >> 1)) / interHeight;

  for (Int i = 0; i < interWidth; i++)
  {
    Pel *pSrc = &pSrcY[i];
    Pel *pDst = &tempY[i];
    for (Int j = 0; j < interHeight; j++)
    {
      Int iRefPos16 = ((j * iScaleY + iAddY) >> iShiftM4) - iDeltaY;
      Int iPhase = iRefPos16  & 15;
      Int iRefPos = iRefPos16 >> 4;
      Int temp = 0;

      for (Int k = 0; k < NTAPS_US_CHROMA; k++)
      {
        Int m = Clip3( 0, srcHeight - 1, iRefPos + k - iTapOffset);
        temp += filterVer[iPhase][k] * pSrc[m * srcStride];
      }
      pDst[j * interStride] = temp;
    }
  }

  //===== horizontal upsampling =====

  //iAdd = (((interWidth * iPhaseY) << (iShift - 2)) + (dstWidth >> 1)) / dstWidth + (1 << (iShift - 5));
  //iScale = ((interWidth << iShift) + (dstWidth >> 1)) / dstWidth;
   
  Int offset = 1 << (2 * US_FILTER_PREC - 1);
  Int shift = 2 * US_FILTER_PREC; 
  for (Int j = 0; j < dstHeight; j++)
  {
    Pel *pSrc = &tempY[j * interStride];
    Pel *pDst = &pDstY[j * dstStride];

    for (Int i = 0; i < dstWidth; i++)
    {
      int iRefPos16 = ((i * iScaleX + iAddX) >> iShiftM4) - iDeltaX;
      int iPhase = iRefPos16 & 15;
      int iRefPos = iRefPos16 >> 4;
      Int temp = 0;

      for (Int k = 0; k < NTAPS_US_CHROMA; k++)
      {
        Int m = Clip3(0, interWidth - 1, iRefPos + k - iTapOffset);
        temp += filterHor[iPhase][k] * pSrc[m];
      }
      pDst[i] = Clip3(0, 255, (temp + offset) >> shift);
    }
  }
}
#else
Void TEncAuf::xDeriveFilterCoef(Int dir)
{
  for (Int filterIdx = 0 ; filterIdx < FILTER_SET; filterIdx++)
  {
    if (m_iCount[dir][filterIdx])
    {
      gnsSolveByChol(m_dAutoCorr[dir][filterIdx], m_dCrosCorr[dir][filterIdx], m_dCoeff[dir][filterIdx], NTAPS_US_LUMA);
    }
  }
}

void TEncAuf::xQuantizeCoeff(Int dir, Int iSliceInterval)
{
  Int filterIdx;

  for ( filterIdx = 1 ; filterIdx < 8; filterIdx++)
  {
    if (m_iCount[dir][filterIdx])
    {
      Int sum = 0;
      for (int i = 0; i < NTAPS_US_LUMA; i++ )
      {
        if(i == NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16)
        {
          continue;
        }
        double sign = m_dCoeff[dir][filterIdx][i] > 0.0 ? 1.0: -1.0;
        m_dCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign))/(double)(1<<US_FILTER_PREC);

        m_iCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign));
        sum +=  m_iCoeff[dir][filterIdx][i];
      }
      m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16] = (1<<US_FILTER_PREC) - sum;
      m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16] / (double)(1<<US_FILTER_PREC);
    }
    else
    {
      for (int i = 0; i < NTAPS_US_LUMA; i++ )
      {
        m_iCoeff[dir][filterIdx][i] = usf_luma_fixed[filterIdx][i];
      }
    }
  }

  for ( filterIdx = 1 ; filterIdx < 8; filterIdx++)
  {
    if (m_iCount[dir][filterIdx])
    {
      for (int i = 0; i < NTAPS_US_LUMA; i++ )
      {
        m_iCoeff[dir][16-filterIdx][i] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] ;
        m_dCoeff[dir][16-filterIdx][i] = m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i];
      }
    }
  }

  filterIdx = 0;
  if (m_iCount[dir][filterIdx])
  {
    Int sum = 0;
    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-1] = 0;
    for (int i = 0; i < NTAPS_US_LUMA/2 - 1; i++ )
    {
      double sign = (m_dCoeff[dir][filterIdx][i]+m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i]) > 0.0 ? 1.0: -1.0;
      m_dCoeff[dir][filterIdx][i] = ((int)((m_dCoeff[dir][filterIdx][i] + m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i] ) * (1<<US_FILTER_PREC)/2.0+0.5*sign))/(double)(1<<US_FILTER_PREC);
      m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i] = m_dCoeff[dir][filterIdx][i];

      m_iCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign));
      m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i] = m_iCoeff[dir][filterIdx][i];
      sum +=  m_iCoeff[dir][filterIdx][i]*2;
    }
    m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = (1<<US_FILTER_PREC) - sum;
    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] / (double)(1<<US_FILTER_PREC);
  }
  else
  {
    for (int i = 0; i < NTAPS_US_LUMA; i++ )
    {
      m_iCoeff[dir][filterIdx][i] = usf_luma_fixed[filterIdx][i];
    }
  }

  filterIdx = 8;
  if (m_iCount[dir][filterIdx])
  {
    Int sum = 0;
    for (int i = 0; i < NTAPS_US_LUMA/2 - 1; i++ )
    {
      double sign = (m_dCoeff[dir][filterIdx][i]+m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i]) > 0.0 ? 1.0: -1.0;
      m_dCoeff[dir][filterIdx][i] = ((int)((m_dCoeff[dir][filterIdx][i] + m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] ) * (1<<US_FILTER_PREC)/2+0.5*sign))/(double)(1<<US_FILTER_PREC);
      m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] = m_dCoeff[dir][filterIdx][i];

      m_iCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign));
      m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] = m_iCoeff[dir][filterIdx][i];
      sum +=  m_iCoeff[dir][filterIdx][i]*2;
    }
    m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = ((1<<US_FILTER_PREC) - sum)/2;
    m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1];

    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] / (double)(1<<US_FILTER_PREC);
    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 ] = m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1];

  }
  else
  {
    for (int i = 0; i < NTAPS_US_LUMA; i++ )
    {
      m_iCoeff[dir][filterIdx][i] = usf_luma_fixed[filterIdx][i];
    }
  }

  Int coeff_threshold = 62;
  if (iSliceInterval == 1)
    coeff_threshold += 10;

  if (m_iCoeff[dir][0][NTAPS_US_LUMA/2 -1] >= coeff_threshold)
  {
    for( Int j = 0; j < 16; j++)
    {
      memcpy( &m_iCoeff[dir][j][0], &usf_luma_fixed[j][0], sizeof(Int) * NTAPS_US_LUMA );
    }
  }
}

Bool TEncAuf::xSelectFilter( TComPicYuv *pPicYuvOrg, TComPicYuv *pPicYuvRecBase)
{
  Int sumHAD_fix = 0;
  Int sumHAD_auf = 0;

  Pel* pOrgY   = pPicYuvOrg->getLumaAddr();
  Int orgWidth = pPicYuvOrg->getWidth() - pPicYuvOrg->getPicCropLeftOffset() - pPicYuvOrg->getPicCropRightOffset();
  Int orgHeight = pPicYuvOrg->getHeight() - pPicYuvOrg->getPicCropTopOffset() - pPicYuvOrg->getPicCropBottomOffset();
  Int orgStride = pPicYuvOrg->getStride();
  Pel* pBaseY    = pPicYuvRecBase->getLumaAddr();
  Int baseWidth  = pPicYuvRecBase->getWidth() - pPicYuvRecBase->getPicCropLeftOffset() - pPicYuvRecBase->getPicCropRightOffset();
  Int baseHeight = pPicYuvRecBase->getHeight() - pPicYuvRecBase->getPicCropTopOffset() - pPicYuvRecBase->getPicCropBottomOffset();
  Int baseStride = pPicYuvRecBase->getStride();

  Pel* pDstY = m_pPicYuvTemp[0]->getLumaAddr();
  Int  dstStride = m_pPicYuvTemp[0]->getStride();
  Pel* tempY = m_pPicYuvTemp[1]->getLumaAddr();

  xUpsampleTwoD( pDstY, tempY, orgWidth, orgHeight, dstStride, pBaseY, baseWidth, baseHeight, baseStride, usf_luma_fixed, usf_luma_fixed );
  sumHAD_fix = m_pcRdCost->calcHAD         ( pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight );

  xUpsampleTwoD( pDstY, tempY, orgWidth, orgHeight, dstStride, pBaseY, baseWidth, baseHeight, baseStride, m_iCoeff[1], m_iCoeff[0] );
  sumHAD_auf = m_pcRdCost->calcHAD         ( pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight );

  if( sumHAD_auf >= sumHAD_fix )
  {
    return false;
  }
  else
  {
    return true;
  }
}
#endif

#if IL_AUF
void TEncAuf::xUpsampleTwoD(Pel *pDstY, Pel *tempY, Int dstWidth, Int dstHeight, Int dstStride, Pel *pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int filterVer[16][8], Int filterHor[16][8], Int iColor)
#else
void TEncAuf::xUpsampleTwoD(Pel* pDstY, Pel* tempY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int filterVer[16][8], Int filterHor[16][8] )
#endif
{
  Int   iShift = 16;
  Int   iPhaseY = 0;
  Int   iShiftM4 = iShift - 4;

  Int   iDelta     = 4 * iPhaseY; 

  Int   iAdd, iScale;

  Int interWidth = srcWidth;
  Int interStride = dstStride;
  Int interHeight = dstHeight;
#if IL_AUF
  Int iFilterTaps = (iColor == COLOR_Y) ? NTAPS_US_LUMA : NTAPS_US_CHROMA;
#else
  Int iFilterTaps = NTAPS_US_LUMA;
#endif
  Int iTapOffset = iFilterTaps / 2 - 1;

  //===== vertical upsampling =====
  iAdd   = ( ( ( srcHeight * iPhaseY ) << ( iShift - 2 ) ) + ( interHeight >> 1 ) ) / interHeight+ ( 1 << ( iShift - 5 ) );

  iScale = ( ( srcHeight << iShift ) + ( interHeight >> 1 ) ) / interHeight;

  for (Int i = 0; i < interWidth; i++ )
  {
    Pel *pSrc = &pSrcY[i];
    Pel *pDst = &tempY[i];
    for (Int j = 0; j < interHeight; j++)
    {
      Int iRefPos16 = (( j * iScale + iAdd ) >> iShiftM4 ) - iDelta;
      Int iPhase    = iRefPos16  & 15;
      int iRefPos   = iRefPos16 >>  4;

      Int temp = 0;

      for (Int k = 0; k < iFilterTaps; k++)
      {
        Int m = Clip3(0, srcHeight - 1, iRefPos + k - iTapOffset);
        temp += filterVer[iPhase][k] * pSrc[m * srcStride];
      }
      pDst[j * interStride] = temp;
    }
  }

  //===== horizontal upsampling =====

  iAdd   = ( ( ( interWidth * iPhaseY ) << ( iShift - 2 ) ) + ( dstWidth >> 1 ) ) / dstWidth + ( 1 << ( iShift - 5 ) );
  iScale = ( ( interWidth << iShift ) + ( dstWidth >> 1 ) ) / dstWidth;
   
  Int offset = 1<<(2*US_FILTER_PREC-1);
  Int shift = 2*US_FILTER_PREC; 
  for( int j = 0; j < dstHeight; j++ )
  {
    Pel* pSrc = &tempY[j*interStride];
    Pel* pDst = &pDstY[j*dstStride];

    for( int i = 0; i < dstWidth; i++ )
    {
      int iRefPos16 = (( i * iScale + iAdd ) >> iShiftM4 ) - iDelta;
      int iPhase    = iRefPos16  & 15;
      int iRefPos   = iRefPos16 >>  4;

      Int temp = 0;

      for (Int k = 0; k < iFilterTaps; k++ )
      {
        Int m = Clip3(0, interWidth - 1, iRefPos + k - iTapOffset);
        temp += filterHor[iPhase][k] * pSrc[m];
      }
      pDst[i] = Clip3(0, 255, ( temp + offset ) >> shift );
    }
  }
}


Void TEncAuf::smoothOrgPic(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Pel* tempFrameBuf )
{
  assert (dstWidth == srcWidth && dstHeight == srcHeight);

  Int fitler_smooth[10] = { 2,     0,   -14,    33,    86,    33,   -14,     0,     2,     0 };

  Int *filter = &fitler_smooth[0];
  
  //vertical filter
  for( int i = 0; i < dstWidth; i++ )
  {
    Pel* pSrc = &pSrcY[i];
    Pel* pDst = &tempFrameBuf[i];
    for( int j = 0; j < dstHeight; j++ )
    {
      int temp = 0;
      for( int k = 0; k < 10; k++ )
      {
        int m = Clip3( 0, srcHeight - 1, j + k - 4);
        temp += filter[k] * pSrc[m*srcStride];
      }
      pDst[j*dstStride] = Clip3(0, 255, ( temp + 64 ) >> 7 );
    }
  }

  //===== horizontal filter =====
  for( int j = 0; j < dstHeight; j++ )
  {
    Pel* pSrc = &tempFrameBuf[j*dstStride];
    Pel* pDst = &pDstY[j*dstStride];

    for( int i = 0; i < dstWidth; i++ )
    {
      int temp = 0;
      for( int k = 0; k < 10; k++ )
      {
        int m = Clip3( 0, srcWidth - 1, i + k - 4 );
        temp += filter[k] * pSrc[m];
      }
      pDst[i] = Clip3( 0, 255, ( temp + 64 ) >> 7);
    }
  }
}

void TEncAuf::xDownsampleOneD(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int dir)
{
  // jdong
  int filter16[2][16][12] =  // 0.9pi
  {
    {
      { 0,     5,    -6,   -10,    37,    76,    37,   -10,    -6,     5,     0,     0},
      { 0,     5,    -4,   -11,    33,    76,    40,    -9,    -7,     5,     0,     0},
      {-1,     5,    -3,   -12,    29,    75,    45,    -7,    -8,     5,     0,     0},
      {-1,     4,    -2,   -13,    25,    75,    48,    -5,    -9,     5,     1,     0},
      {-1,     4,    -1,   -13,    22,    73,    52,    -3,   -10,     4,     1,     0},
      {-1,     4,     0,   -13,    18,    72,    55,    -1,   -11,     4,     2,    -1},
      {-1,     4,     1,   -13,    14,    70,    59,     2,   -12,     3,     2,    -1},
      {-1,     3,     1,   -13,    11,    68,    62,     5,   -12,     3,     2,    -1},
      {-1,     3,     2,   -13,     8,    65,    65,     8,   -13,     2,     3,    -1},
      {-1,     2,     3,   -12,     5,    62,    68,    11,   -13,     1,     3,    -1},
      {-1,     2,     3,   -12,     2,    59,    70,    14,   -13,     1,     4,    -1},
      {-1,     2,     4,   -11,    -1,    55,    72,    18,   -13,     0,     4,    -1},
      { 0,     1,     4,   -10,    -3,    52,    73,    22,   -13,    -1,     4,    -1},
      { 0,     1,     5,    -9,    -5,    48,    75,    25,   -13,    -2,     4,    -1},
      { 0,     1,     5,    -8,    -7,    44,    75,    28,   -12,    -3,     5,     0},
      { 0,     0,     5,    -7,    -9,    40,    76,    33,   -11,    -4,     5,     0}
    },
    {
      {2,    -3,    -9,     6,    39,    58,    39,     6,    -9,    -3,     2,     0},   
      {2,    -3,    -9,     4,    38,    58,    43,     7,    -9,    -4,     1,     0},   
      {2,    -2,    -9,     2,    35,    58,    44,     9,    -8,    -4,     1,     0},   
      {1,    -2,    -9,     1,    34,    58,    46,    11,    -8,    -5,     1,     0},   
      {1,    -1,    -8,    -1,    31,    57,    47,    13,    -7,    -5,     1,     0},   
      {1,    -1,    -8,    -2,    29,    56,    49,    15,    -7,    -6,     1,     1},   
      {1,     0,    -8,    -3,    26,    55,    51,    17,    -7,    -6,     1,     1},   
      {1,     0,    -7,    -4,    24,    54,    52,    19,    -6,    -7,     1,     1},   
      {1,     0,    -7,    -5,    22,    53,    53,    22,    -5,    -7,     0,     1},   
      {1,     1,    -7,    -6,    19,    52,    54,    24,    -4,    -7,     0,     1},   
      {1,     1,    -6,    -7,    17,    51,    55,    26,    -3,    -8,     0,     1},   
      {1,     1,    -6,    -7,    15,    49,    56,    29,    -2,    -8,    -1,     1},   
      {0,     1,    -5,    -8,    13,    48,    57,    31,    -1,    -8,    -1,     1},   
      {0,     1,    -5,    -8,    11,    46,    58,    34,     1,    -9,    -2,     1},   
      {0,     1,    -4,    -8,     9,    44,    58,    35,     2,    -9,    -2,     2},   
      {0,     1,    -4,    -9,     7,    43,    58,    38,     4,    -9,    -3,     2}
    }
  };

  int iFilter;

  if (dir)
  {
    if (  1 * srcHeight >  2 * dstHeight )   
    {
      assert(0);
    }
    else if (  3 * srcHeight >  5 * dstHeight ) 
    {
      iFilter  = 1;
    }
    else if (  4 * srcHeight >  5 * dstHeight )   
    {
      iFilter  = 0;
    }
    else
    {
      assert (0);
    }
  }
  else
  {
    if (  1 * srcWidth >  2 * dstWidth )   
    {
      assert(0);
    }
    else if (  3 * srcWidth >  5 * dstWidth ) 
    {
      iFilter  = 1;
    }
    else if (  4 * srcWidth >  5 * dstWidth )   
    {
      iFilter  = 0;
    }
    else
    {
      assert (0);
    }
  }
  Int   iShift = 16;
  Int   iPhaseY = 0;
  Int   iShiftM4 = iShift - 4;

  Int   iDelta     = 4*iPhaseY;


  Int   iAdd, iScale;
  if (dir) // ver
  {
    assert(dstWidth == srcWidth);

    iAdd   = ( ( ( srcHeight * iPhaseY ) << ( iShift - 2 ) ) + ( dstHeight >> 1 ) ) / dstHeight+ ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcHeight << iShift ) + ( dstHeight >> 1 ) ) / dstHeight;
  }
  else //hor
  {
    assert(dstHeight == srcHeight);

    iAdd   = ( ( ( srcWidth * iPhaseY ) << ( iShift - 2 ) ) + ( dstWidth >> 1 ) ) / dstWidth + ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcWidth << iShift ) + ( dstWidth >> 1 ) ) / dstWidth;
  }

  if (dir) //===== vertical downsampling =====
  {
    for( int i = 0; i < dstWidth; i++ )
    {
      Pel* pSrc = &pSrcY[i];
      Pel* pDst = &pDstY[i];
      for( int j = 0; j < dstHeight; j++ )
      {
        int iRefPos16 = (( j * iScale + iAdd ) >> iShiftM4 ) - iDelta;
        int iPhase    = iRefPos16  & 15;
        int iRefPos   = iRefPos16 >>  4;

        int temp = 0;
        for( int k = 0; k < 12; k++ )
        {
          int m = Clip3(0, srcHeight - 1, iRefPos + k - 5 );
          temp += filter16[iFilter][iPhase][k] * pSrc[m*srcStride];
        }

        pDst[j*dstStride] = Clip3(0, 255, ( temp + 64 ) >> 7 );
      }
    }
  } 
  else//===== horizontal downsampling =====
  {
    for( int j = 0; j < dstHeight; j++ )
    {
      Pel* pSrc = &pSrcY[j*srcStride];
      Pel* pDst = &pDstY[j*dstStride];

      for( int i = 0; i < dstWidth; i++ )
      {
        int iRefPos16 = (( i * iScale + iAdd ) >> iShiftM4 ) - iDelta;
        int iPhase    = iRefPos16  & 15;
        int iRefPos   = iRefPos16 >>  4;

        int temp = 0;
        for( int k = 0; k < 12; k++ )
        {
          int m = Clip3(0, srcWidth - 1, iRefPos + k - 5 );
          temp += filter16[iFilter][iPhase][k] * pSrc[m];
        }
        pDst[i] = Clip3( 0, 255, ( temp + 64 ) >> 7 );
      }
    }
  }
}

#if IL_AUF
void TEncAuf::xUpsampleOneD(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, 
                            Int srcHeight, Int srcStride, Int dir, Int iColor)
#else
void TEncAuf::xUpsampleOneD(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, 
                            Int srcHeight, Int srcStride, Int dir)
#endif
{
  Int iShift = 16;
  Int iPhaseY = 0;
  Int iShiftM4 = iShift - 4;
  Int iDelta = 4 * iPhaseY;
#if IL_AUF
  Int iUpSampleTaps = (iColor == COLOR_Y) ? NTAPS_US_LUMA : NTAPS_US_CHROMA;
#else
  Int iUpSampleTaps = NTAPS_US_LUMA;
#endif
  Int iTapOffset = iUpSampleTaps / 2 - 1;

  Int   iAdd, iScale;
  if (dir) // ver
  {
    assert(dstWidth == srcWidth);
    iAdd   = ( ( ( srcHeight * iPhaseY ) << ( iShift - 2 ) ) + ( dstHeight >> 1 ) ) / dstHeight+ ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcHeight << iShift ) + ( dstHeight >> 1 ) ) / dstHeight;
  }
  else //hor
  {
    assert(dstHeight == srcHeight);
    iAdd   = ( ( ( srcWidth * iPhaseY ) << ( iShift - 2 ) ) + ( dstWidth >> 1 ) ) / dstWidth + ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcWidth << iShift ) + ( dstWidth >> 1 ) ) / dstWidth;
  }

  if (dir) //===== vertical upsampling =====
  {
    for (Int i = 0; i < dstWidth; i++)
    {
      Pel *pSrc = &pSrcY[i];
      Pel *pDst = &pDstY[i];
      for (Int j = 0; j < dstHeight; j++ )
      {
        Int iRefPos16 = (( j * iScale + iAdd ) >> iShiftM4 ) - iDelta;
        Int iPhase    = iRefPos16  & 15;
        Int iRefPos   = iRefPos16 >>  4;

        Int temp = 0;
        Int offset = 1<<(US_FILTER_PREC-1);
        for (Int k = 0; k < iUpSampleTaps; k++)
        {
          Int m = Clip3(0, srcHeight - 1, iRefPos + k - iTapOffset);
#if IL_AUF
          temp += m_iCoeff[iColor][dir][iPhase][k] * pSrc[m * srcStride];
#else
          temp += m_iCoeff[dir][iPhase][k] * pSrc[m*srcStride];
#endif
        }
        pDst[j * dstStride] = Clip3(0, 255, ( temp + offset ) >> US_FILTER_PREC);  ///< clip is not necessary
      }
    }
  } 
  else//===== horizontal upsampling =====
  {
    for (Int j = 0; j < dstHeight; j++)
    {
      Pel *pSrc = &pSrcY[j * srcStride];
      Pel *pDst = &pDstY[j * dstStride];

      for (Int i = 0; i < dstWidth; i++)
      {
        Int iRefPos16 = (( i * iScale + iAdd ) >> iShiftM4) - iDelta;
        Int iPhase = iRefPos16 & 15;
        Int iRefPos = iRefPos16 >> 4;

        Int temp = 0;
        Int offset = 1 << (US_FILTER_PREC - 1);
        for (Int k = 0; k < iUpSampleTaps; k++)
        {
          Int m = Clip3(0, srcWidth - 1, iRefPos + k - iTapOffset);
#if IL_AUF
          temp += m_iCoeff[iColor][dir][iPhase][k] * pSrc[m];
#else
          temp += m_iCoeff[dir][iPhase][k] * pSrc[m];
#endif
        }
        pDst[i] = Clip3(0, 255, (temp + offset) >> US_FILTER_PREC);  ///< clip is not necessary
      }
    }
  }
}

#if IL_AUF
Void TEncAuf::xGetStatistics(Pel* pOrgY, Int orgWidth, Int orgHeight, Int orgStride, Pel* pRefY, 
                             Int refWidth, Int refHeight, Int refStride, Int dir, COLORType eColor, TComPicYuv *pPicMap)
#else
Void TEncAuf::xGetStatistics(Pel* pOrgY, Int orgWidth, Int orgHeight, Int orgStride, Pel* pRefY, 
                             Int refWidth, Int refHeight, Int refStride, Int dir)
#endif
{
  //initilize 
  Int idx;
  Int l, k; 
  Int x, y;

#if IL_AUF
  Pel *pMap = NULL;
  assert(pPicMap);
  Int iMapStride;
  switch (eColor)
  {
  case COLOR_Y:
    pMap = pPicMap->getLumaAddr();
    iMapStride = pPicMap->getStride();
    break;
  case COLOR_Cb:
    pMap = pPicMap->getCbAddr();
    iMapStride = pPicMap->getCStride();
    break;
  case COLOR_Cr:
    pMap = pPicMap->getCrAddr();
    iMapStride = pPicMap->getCStride();
    break;
  default:
    assert(0);
  }

  Int offset = (eColor == COLOR_Y) ? NTAPS_US_LUMA / 2 - 1 : NTAPS_US_CHROMA / 2 - 1;
  Int iFilterTap = (eColor == COLOR_Y) ? NTAPS_US_LUMA : NTAPS_US_CHROMA;
#else
  Int offset = NTAPS_US_LUMA/2 - 1;
  Int iFilterTap = NTAPS_US_LUMA;
#endif
  Int   iShift = 16;
  Int   iPhase = 0;

  Int   iShiftM4 = iShift - 4;

  Int   iDelta     = 4*iPhase;

  Int   iAdd, iScale;
  if (dir) // ver
  {
    assert(refWidth == orgWidth);

    iAdd   = ( ( ( refHeight * iPhase ) << ( iShift - 2 ) ) + ( orgHeight >> 1 ) ) / orgHeight+ ( 1 << ( iShift - 5 ) );
    iScale = ( ( refHeight << iShift ) + ( orgHeight >> 1 ) ) / orgHeight;
  }
  else //hor
  {
    assert(refHeight == orgHeight);

    iAdd   = ( ( ( refWidth * iPhase ) << ( iShift - 2 ) ) + ( orgWidth >> 1 ) ) / orgWidth + ( 1 << ( iShift - 5 ) );
    iScale = ( ( refWidth << iShift ) + ( orgWidth >> 1 ) ) / orgWidth;
  }

  for(idx = 0; idx < FILTER_SET; idx++)
  {
    for(l = 0; l < iFilterTap; l++)
    {
      for(k = 0; k < iFilterTap; k++)
      {
#if IL_AUF
        m_dAutoCorr[eColor][dir][idx][l][k] = 0.0;
#else
        m_dAutoCorr[dir][idx][l][k] = 0.0;
#endif
      }
      
#if IL_AUF
      m_dCrosCorr[eColor][dir][idx][l] = 0.0;
#else
      m_dCrosCorr[dir][idx][l] = 0.0;
#endif
    }
#if IL_AUF
    m_dPixAcc[eColor][dir][idx] = 0.0;
    m_iCount[eColor][dir][idx] = 0;
#else
    m_dPixAcc[dir][idx] = 0.0;
    m_iCount[dir][idx] = 0;
#endif
  }

  if (dir) // ver
  {
    for (y = offset; y < orgHeight - offset - 1; y++)
    {
      Pel* org = pOrgY + y * orgStride + offset;
#if IL_AUF
      Pel *pM = pMap + y * iMapStride + offset;
#endif

      Int iRefPos16 = ((y*iScale + iAdd) >> iShiftM4) - iDelta;
      Int iRefPhase = iRefPos16 & 15;
      Int iRefPos   = iRefPos16 >> 4;

      Pel* ref = pRefY + iRefPos * refStride + offset;

      for (x = offset; x < orgWidth - offset - 1; x++)
      {
        for(l = 0; l < iFilterTap; l++)
        {
          int filter_pos_y1 = l - offset;

          for(k = 0; k < iFilterTap; k++)			
          {
            int filter_pos_y2 = k - offset ;

            double ac = (*(ref + filter_pos_y1*refStride ))*(*(ref+ filter_pos_y2*refStride ));
#if IL_AUF
            if (*pM)
            {
              m_dAutoCorr[eColor][dir][iRefPhase][l][k] += ac;
            }
#else
            m_dAutoCorr[dir][iRefPhase][l][k] += ac;
#endif
          }

          double cc = (*org) * (*(ref+ filter_pos_y1*refStride ));
#if IL_AUF
          if (*pM)
          {
            m_dCrosCorr[eColor][dir][iRefPhase][l] += cc;
          }
#else
          m_dCrosCorr[dir][iRefPhase][l] += cc;
#endif
        }
#if IL_AUF
        m_dPixAcc[eColor][dir][iRefPhase] += (*org)*(*org); 
        if (*pM)
        {
          m_iCount[eColor][dir][iRefPhase] ++;
        }
#else
        m_dPixAcc[dir][iRefPhase] += (*org)*(*org);
        m_iCount[dir][iRefPhase] ++;
#endif
        org ++;
        ref ++;
#if IL_AUF
        pM++;
#endif
      }
    }
  } 
  else // hor
  {
    for (x = offset; x < orgWidth - offset - 1; x++)
    {
      Pel* org = pOrgY + x + orgStride * offset;
#if IL_AUF
      Pel *pM = pMap + x + iMapStride * offset;
#endif
      Int iRefPos16 = ((x*iScale + iAdd) >> iShiftM4) - iDelta;
      Int iRefPhase = iRefPos16 & 15;
      Int iRefPos   = iRefPos16 >> 4;

      Pel* ref = pRefY + iRefPos + refStride * offset;
      for (y = offset; y < orgHeight - offset - 1; y++)
      {
        for(l = 0; l < iFilterTap; l++)
        {
          int filter_pos_x1 = l - offset;

          for(k = 0; k < iFilterTap; k++)			
          {
            int filter_pos_x2 = k - offset ;

            double ac = (*(ref + filter_pos_x1 ))*(*(ref+ filter_pos_x2 ));
#if IL_AUF
        if (*pM)
        {
            m_dAutoCorr[eColor][dir][iRefPhase][l][k] += ac;
        }
#else
            m_dAutoCorr[dir][iRefPhase][l][k] += ac;
#endif
          }

          double cc = (*org) * (*(ref+ filter_pos_x1 ));
#if IL_AUF
        if (*pM)
        {
          m_dCrosCorr[eColor][dir][iRefPhase][l] += cc;
        }
#else
          m_dCrosCorr[dir][iRefPhase][l] += cc;
#endif
        }
#if IL_AUF
        m_dPixAcc[eColor][dir][iRefPhase] += (*org)*(*org); 
        if (*pM)
        {
          m_iCount[eColor][dir][iRefPhase] ++;
        }
#else
        m_dPixAcc[dir][iRefPhase] += (*org)*(*org); 
        m_iCount[dir][iRefPhase] ++;
#endif
        org +=orgStride;
        ref +=refStride;
#if IL_AUF
        pM += iMapStride;
#endif
      }
    }
  }

  double backupAC[16][NTAPS_US_LUMA][NTAPS_US_LUMA];
  double backupCC[16][NTAPS_US_LUMA];
  double backupPixelACC[16];

  for (idx = 1 ; idx < 16; idx++ )
  {
    for(l = 0; l < iFilterTap; l++)
    {
      for(k = 0; k < iFilterTap; k++)			
      {
#if IL_AUF
        backupAC[idx][l][k] = m_dAutoCorr[eColor][dir][idx][l][k];
#else
        backupAC[idx][l][k] = m_dAutoCorr[dir][idx][l][k];
#endif
      }
#if IL_AUF
      backupCC[idx][l] = m_dCrosCorr[eColor][dir][idx][l];
#else
      backupCC[idx][l] = m_dCrosCorr[dir][idx][l];
#endif
    }
#if IL_AUF
    backupPixelACC[idx] = m_dPixAcc[eColor][dir][idx];
#else
    backupPixelACC[idx] = m_dPixAcc[dir][idx];
#endif
  }

  for ( idx = 1 ; idx < 16; idx++)
  {
#if IL_AUF
    if (m_iCount[eColor][dir][idx] && m_iCount[eColor][dir][16-idx] )
#else
    if (m_iCount[dir][idx] && m_iCount[dir][16-idx] )
#endif
    {
      for(l = 0; l < iFilterTap; l++)
      {
        for(k = 0; k < iFilterTap; k++)			
        {
#if IL_AUF
          m_dAutoCorr[eColor][dir][idx][l][k] = backupAC[idx][l][k] + backupAC[16-idx][iFilterTap - 1 - l][iFilterTap - 1 - k];
#else
          m_dAutoCorr[dir][idx][l][k] = backupAC[idx][l][k] + backupAC[16-idx][NTAPS_US_LUMA-1-l][NTAPS_US_LUMA-1-k];
#endif
        }
#if IL_AUF
        m_dCrosCorr[eColor][dir][idx][l] = backupCC[idx][l] + backupCC[16-idx][iFilterTap - 1 - l];
        m_dPixAcc[eColor][dir][idx] = backupPixelACC[dir] + backupPixelACC[16 -dir];
#else
        m_dCrosCorr[dir][idx][l] = backupCC[idx][l] + backupCC[16-idx][NTAPS_US_LUMA-1-l];
        m_dPixAcc[dir][idx] = backupPixelACC[dir] + backupPixelACC[16 -dir];
#endif
      }
    }
  }

  idx = 0;
  for(l = 0; l < iFilterTap; l++)
  {
    for(k = 0; k < iFilterTap; k++)			
    {
#if IL_AUF
      backupAC[idx][l][k] = m_dAutoCorr[eColor][dir][idx][l][k];
#else
      backupAC[idx][l][k] = m_dAutoCorr[dir][idx][l][k];
#endif
    }
#if IL_AUF
    backupCC[idx][l] = m_dCrosCorr[eColor][dir][idx][l];
#else
    backupCC[idx][l] = m_dCrosCorr[dir][idx][l];
#endif
  }
#if IL_AUF
  backupPixelACC[idx] = m_dPixAcc[eColor][dir][idx];
  if (m_iCount[eColor][dir][idx] )
#else
  backupPixelACC[idx] = m_dPixAcc[dir][idx];
  if (m_iCount[dir][idx] )
#endif
  {
    for(l = 0; l < iFilterTap - 1; l++)
    {
      for(k = 0; k < iFilterTap - 1; k++)			
      {
#if IL_AUF
        m_dAutoCorr[eColor][dir][idx][l][k] = backupAC[idx][l][k] + backupAC[idx][iFilterTap - 2 - l][iFilterTap - 2 - k];
#else
        m_dAutoCorr[dir][idx][l][k] = backupAC[idx][l][k] + backupAC[idx][NTAPS_US_LUMA-2-l][NTAPS_US_LUMA-2-k];
#endif
      }
#if IL_AUF
      m_dCrosCorr[eColor][dir][idx][l] = backupCC[idx][l] + backupCC[idx][iFilterTap - 2 - l];
      m_dPixAcc[eColor][dir][idx] = backupPixelACC[dir] * 2;
#else
      m_dCrosCorr[dir][idx][l] = backupCC[idx][l] + backupCC[idx][NTAPS_US_LUMA-2-l];
      m_dPixAcc[dir][idx] = backupPixelACC[dir]*2;
#endif
    }
  }


#if IL_AUF
  m_dCrosCorr[eColor][dir][0][iFilterTap - 1]  = 0.0;
#else
  m_dCrosCorr[dir][0][NTAPS_US_LUMA-1]  = 0.0;
#endif
  for(l = 0; l < iFilterTap; l++)
  {
    for(k = 0; k < iFilterTap; k++)			
    {
      if (k == iFilterTap - 1 || l == iFilterTap - 1 )
      {
#if IL_AUF
        m_dAutoCorr[eColor][dir][0][l][k] = 0.0;
#else
        m_dAutoCorr[dir][0][l][k] = 0.0;
#endif
      }
    }
  }
}


#define ROUND(a)  (((a) < 0)? (int)((a) - 0.5) : (int)((a) + 0.5))
#define REG              0.0001
#define REG_SQR          0.0000001

//Find filter coeff related
Int TEncAuf::gnsCholeskyDec(Double inpMatr[NTAPS_US_LUMA][NTAPS_US_LUMA], Double outMatr[NTAPS_US_LUMA][NTAPS_US_LUMA], Int noEq)
{ 
  Int i, j, k;     /* Looping Variables */
  Double scale;       /* scaling factor for each row */
  Double invDiag[NTAPS_US_LUMA];  /* Vector of the inverse of diagonal entries of outMatr */

  //  Cholesky decomposition starts

  for(i = 0; i < noEq; i++)
  {
    for(j = i; j < noEq; j++)
    {
      /* Compute the scaling factor */
      scale = inpMatr[i][j];
      if ( i > 0) 
      {
        for( k = i - 1 ; k >= 0 ; k--)
        {
          scale -= outMatr[k][j] * outMatr[k][i];
        }
      }
      /* Compute i'th row of outMatr */
      if(i == j)
      {
        if(scale <= REG_SQR ) // if(scale <= 0 )  /* If inpMatr is singular */
        {
          return 0;
        }
        else
        {
          /* Normal operation */
          invDiag[i] =  1.0 / (outMatr[i][i] = sqrt(scale));
        }
      }
      else
      {
        outMatr[i][j] = scale * invDiag[i]; /* Upper triangular part          */
        outMatr[j][i] = 0.0;              /* Lower triangular part set to 0 */
      }                    
    }
  }
  return 1; /* Signal that Cholesky factorization is successfully performed */
}


Void TEncAuf::gnsTransposeBacksubstitution(Double U[NTAPS_US_LUMA][NTAPS_US_LUMA], Double rhs[], Double x[], Int order)
{
  Int i,j;              /* Looping variables */
  Double sum;              /* Holds backsubstitution from already handled rows */

  /* Backsubstitution starts */
  x[0] = rhs[0] / U[0][0];               /* First row of U'                   */
  for (i = 1; i < order; i++)
  {         /* For the rows 1..order-1           */

    for (j = 0, sum = 0.0; j < i; j++) /* Backsubst already solved unknowns */
    {
      sum += x[j] * U[j][i];
    }
    x[i] = (rhs[i] - sum) / U[i][i];       /* i'th component of solution vect.  */
  }
}

Void  TEncAuf::gnsBacksubstitution(Double R[NTAPS_US_LUMA][NTAPS_US_LUMA], Double z[NTAPS_US_LUMA], Int R_size, Double A[NTAPS_US_LUMA])
{
  Int i, j;
  Double sum;

  R_size--;

  A[R_size] = z[R_size] / R[R_size][R_size];

  for (i = R_size-1; i >= 0; i--)
  {
    for (j = i + 1, sum = 0.0; j <= R_size; j++)
    {
      sum += R[i][j] * A[j];
    }

    A[i] = (z[i] - sum) / R[i][i];
  }
}


Int TEncAuf::gnsSolveByChol(Double LHS[NTAPS_US_LUMA][NTAPS_US_LUMA], Double rhs[NTAPS_US_LUMA], Double x[NTAPS_US_LUMA], Int noEq)
{
  assert(noEq > 0);

  Double aux[NTAPS_US_LUMA];     /* Auxiliary vector */
  Double U[NTAPS_US_LUMA][NTAPS_US_LUMA];    /* Upper triangular Cholesky factor of LHS */
  Int  i, singular;          /* Looping variable */

  /* The equation to be solved is LHSx = rhs */

  /* Compute upper triangular U such that U'*U = LHS */
  if(gnsCholeskyDec(LHS, U, noEq)) /* If Cholesky decomposition has been successful */
  {
    singular = 1;
    /* Now, the equation is  U'*U*x = rhs, where U is upper triangular
    * Solve U'*aux = rhs for aux
    */
    gnsTransposeBacksubstitution(U, rhs, aux, noEq);         

    /* The equation is now U*x = aux, solve it for x (new motion coefficients) */
    gnsBacksubstitution(U, aux, noEq, x);   

  }
  else /* LHS was singular */ 
  {
    singular = 0;

    /* Regularize LHS */
    for(i=0; i < noEq; i++)
    {
      LHS[i][i] += REG;
    }
    /* Compute upper triangular U such that U'*U = regularized LHS */
    singular = gnsCholeskyDec(LHS, U, noEq);
    if ( singular == 1 )
    {
      /* Solve  U'*aux = rhs for aux */  
      gnsTransposeBacksubstitution(U, rhs, aux, noEq);   

      /* Solve U*x = aux for x */
      gnsBacksubstitution(U, aux, noEq, x);      
    }
    else
    {
      x[0] = 1.0;
      for (i = 1; i < noEq; i++ )
      {
        x[i] = 0.0;
      }
    }
  }  
  return singular;
}

#endif // TE4_4_2_1
